/* "comm" match extension for Xtables */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/version.h>
#include <net/sock.h>
#include <net/inet_sock.h>
#include <net/tcp.h>
#include <net/inet_hashtables.h>

#include <linux/netfilter_ipv4.h>
#include <linux/netfilter/x_tables.h>

#include "xt_comm.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("mengyuan");
MODULE_DESCRIPTION("xtables module for comm match");
MODULE_ALIAS("ip6t_comm");
MODULE_ALIAS("ipt_comm");

static bool comm_mt(const struct sk_buff *skb, struct xt_action_param *par)
{
	const struct xt_comm_info *info = par->matchinfo;
	struct task_struct *p;
	struct files_struct *files;
	struct fdtable *fdt;
	struct file *file = NULL;
	struct socket *sock;
	struct socket *lsock = NULL;
	int i, err;
	bool ret = false;
	const struct tcphdr *th;
	struct tcphdr _tcph;
	const struct iphdr *iph;
	struct sock *sk;
	int sdif = inet_sdif(skb);

	if (par->fragoff != 0)
		goto task;
	th = skb_header_pointer(skb, par->thoff, sizeof(_tcph), &_tcph);
	if (th == NULL)
		goto task;
	iph = ip_hdr(skb);
	sk = __inet_lookup_listener(dev_net(skb_dst(skb)->dev), &tcp_hashinfo, skb, __tcp_hdrlen(th),
			iph->saddr, th->source, iph->daddr, ntohs(th->dest), inet_iif(skb), sdif);
//	if (!sk)
//		sk = __inet_lookup_established(dev_net(skb_dst(skb)->dev), &tcp_hashinfo,
//			iph->saddr, th->source, iph->daddr, ntohs(th->dest), inet_iif(skb), sdif);
	if (sk) {
//		printk("MYDBG: sport[%d] dport[%d]", ntohs(th->source), ntohs(th->dest));
		lsock = sk->sk_socket;
	}
	/* TODO: UDP INPUT */
task:
	rcu_read_lock();
	list_for_each_entry_rcu(p, &init_task.tasks, tasks) {
		if (strncmp(p->comm, info->comm, sizeof(p->comm)))
			continue;
		task_lock(p);
		files = p->files;
		if (files) {
			spin_lock(&files->file_lock);
			fdt = files_fdtable(files);
			for (i = 0; i < fdt->max_fds; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
				file = files_lookup_fd_locked(files, i);
#else
				file = fcheck_files(files, i);
#endif
				if (!file)
					continue;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
				sock = sock_from_file(file);
#else
				sock = sock_from_file(file, &err);
#endif
				if (!sock)
					continue;
				if (sock == lsock) {
					ret = true;
                                        break;
				}
				if (!skb || !skb->sk || !skb->sk->sk_socket)
					continue;
				if (sock == skb->sk->sk_socket) {
					ret = true;
					break;
				}
			}
			spin_unlock(&files->file_lock);
		}
		task_unlock(p);
	};
	rcu_read_unlock();

	return ret ^ info->invert;
}

static struct xt_match comm_mt_reg __read_mostly = {
	.name      = "comm",
	.revision  = 0,
	.family    = NFPROTO_UNSPEC,
	.match     = comm_mt,
	.matchsize = sizeof(struct xt_comm_info),
	.hooks     = (1 << NF_INET_LOCAL_IN) | (1 << NF_INET_LOCAL_OUT),
	.me        = THIS_MODULE,
};

static int __init comm_mt_init(void)
{
	return xt_register_match(&comm_mt_reg);
}

static void __exit comm_mt_exit(void)
{
	xt_unregister_match(&comm_mt_reg);
}

module_init(comm_mt_init);
module_exit(comm_mt_exit);
